import { world, system } from '@minecraft/server';
import blockScanner from '../utils/BlockScanner.js';
import PlayerStateManager from '../managers/PlayerStateManager.js';
import { BLOCK_SOUND_MAPPING, SOUND_CONFIGS, DEFAULT_SOUND } from '../utils/Data.js';
import CONFIG from '../config.js';

class CrawlHandler {
    constructor() {
        this.lastTick = new Map();
        this.crawlingStates = new Map();
    }

    // Validation methods
    validateState(player) {
        return PlayerStateManager.getState(player.name)?.crawl ?? null;
    }

    // Main handler method
    handleCrawl(player, currentTick) {
        const lastTick = this.lastTick.get(player.name) ?? -1;
        if (currentTick === lastTick) return;
        this.lastTick.set(player.name, currentTick);

        const crawlState = this.validateState(player);
        if (!crawlState) return;

        const { dimension, location } = player;
        const state = PlayerStateManager.getState(player.name);

        // ONE scan call - gets cached comprehensive data
        const scan = blockScanner.getPlayerScan(player, currentTick);
        const { isInLiquid, hasObstacle, blockGap, canCrawl } = scan;
        const isGapValid = blockGap.exists;

        const isSneaking = player.isSneaking;
        const isOnGround = player.isOnGround;
        const isSprinting = player.isSprinting;
        const velocity = player.getVelocity();
        const isMoving = Math.sqrt(velocity.x ** 2 + velocity.z ** 2) > 0.0001;
        const isAlreadyCrawling = Math.abs(location.y - player.getHeadLocation().y) < 0.301;
        const isSliding = state?.slide?.isSliding || false;
        const isRolling = state?.roll?.isRolling || false;
        const isOnCrawlCooldown = crawlState.lastCrawlTime && (currentTick - crawlState.lastCrawlTime) < CONFIG.CRAWL.COOLDOWN;

        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;
        const forwardX = Math.cos(yawRad);
        const forwardZ = Math.sin(yawRad);
        const movingForward = (velocity.x * forwardX + velocity.z * forwardZ) > 0;

        const wasCrawling = this.crawlingStates.get(player.name) ?? false;
        if (isAlreadyCrawling !== wasCrawling) {
            this.crawlingStates.set(player.name, isAlreadyCrawling);
            if (!isAlreadyCrawling && wasCrawling) {
                this.endCrawl(player, currentTick);
            }
        }

        if (isSneaking && isOnGround && !isSprinting && isMoving && !isInLiquid && 
            !isAlreadyCrawling && !isSliding && !isRolling && state && !hasObstacle && 
            isGapValid && !isOnCrawlCooldown && movingForward) {
            this.startCrawl(player, currentTick, dimension, location);
            this.tryTeleport(player, currentTick, location, yawRad);
        }

        if (currentTick % 100 === 0) this.clearCaches();
    }

    // Start methods
    startCrawl(player, currentTick, dimension, location) {
        const state = PlayerStateManager.getState(player.name);
        if (!state) return;

        state.crawl.isCrawling = true;
        state.crawl.startTime = currentTick;

        PlayerStateManager.setCrawlState(player.name, {
            startTime: currentTick
        });
    }

    // Movement methods
    tryTeleport(player, currentTick, location, yawRad) {
        const setLocation = {
            x: location.x + Math.cos(yawRad) * 0.1,
            y: location.y,
            z: location.z + Math.sin(yawRad) * 0.1
        };
        if (player.tryTeleport(setLocation)) {
            player.teleport(setLocation);
            
            // Use comprehensive scan data for ground block detection
            const scan = blockScanner.getPlayerScan(player, currentTick);
            const blockBelow = scan.ground.block;
            const blockSound = BLOCK_SOUND_MAPPING[blockBelow?.typeId] || DEFAULT_SOUND.sound;
            const volume = 1.5;
            const pitch = 1.0;
            player.dimension.playSound(blockSound, location, { pitch, volume });
            return true;
        }
        return false;
    }

    // End methods
    endCrawl(player, currentTick) {
        const state = PlayerStateManager.getState(player.name);
        if (!state) return;

        state.crawl.isCrawling = false;
        state.crawl.lastCrawlTime = currentTick;
        PlayerStateManager.clearCrawlState(player.name, currentTick);
    }

    // Utility methods
    clearCaches() {
        blockScanner.clearCache();
    }
}

export default CrawlHandler;