import { world, system } from '@minecraft/server';
import blockScanner from '../utils/BlockScanner.js';
import PlayerStateManager from '../managers/PlayerStateManager.js';
import CONFIG from '../config.js';

class RollHandler {
    constructor() {
        this.knockbackIntervals = new Map();
        this.lastTick = new Map();
        this.lastLiquidBlockType = new Map();
        this.fallDamageTimers = new Map();
        this.rollInputTimers = new Map();
        this.rollActiveStates = new Map();
        this.effectTimers = new Map();
    }

    // Validation methods
    validateState(player) {
        return PlayerStateManager.getState(player.name)?.roll ?? null;
    }

    // Warmup methods
    preWarmRoll(player) {
        const rollState = this.validateState(player);
        if (!rollState) return;
        
        try {
            player.playAnimation("animation.humanoid.prewarm");
        } catch (error) {
            console.error(`Pre-warm failed for ${player.name}: ${error}`);
        }
    }

    // Input check methods
    checkRollInputTimeout(player, rollState, currentTick) {
        if (rollState.rollInputReceived && !rollState.isRolling && rollState.rollInputTime !== undefined) {
            const ticksSinceInput = currentTick - rollState.rollInputTime;
            if (ticksSinceInput >= CONFIG.ROLL.DURATION) {
                this.resetFallState(rollState);
                if (this.rollInputTimers.has(player.name)) {
                    system.clearRun(this.rollInputTimers.get(player.name));
                    this.rollInputTimers.delete(player.name);
                }
            } else if (!this.rollInputTimers.has(player.name)) {
                this.manageTimer(player, this.rollInputTimers, () => {
                    if (rollState.rollInputReceived && !rollState.isRolling) {
                        this.resetFallState(rollState);
                    }
                }, CONFIG.ROLL.DURATION - ticksSinceInput);
            }
        }
    }

    // Main handler method
    handleRoll(player, currentTick) {
        const lastTick = this.lastTick.get(player.name) ?? -1;
        if (currentTick === lastTick) return;
        this.lastTick.set(player.name, currentTick);

        const rollState = this.validateState(player);
        if (!rollState) return;

        const { dimension, location } = player;
        
        const scan = blockScanner.getPlayerScan(player, currentTick);
        const { ground, isInLiquid, canRoll } = scan;
        const { block: solidBlockBelow, distance: distanceToSolid } = ground;

        const isOnGround = player.isOnGround;
        const isJumping = player.isJumping;
        const isGliding = player.isGliding;
        const isSneaking = player.isSneaking;
        const isSprinting = player.isSprinting;
        const rollActive = rollState.rollInputReceived || rollState.isRolling;

        this.checkRollInputTimeout(player, rollState, currentTick);

        const lastRollActive = this.rollActiveStates.get(player.name) ?? false;
        if (rollActive && !lastRollActive) {
            dimension.runCommand('gamerule falldamage false');
            this.rollActiveStates.set(player.name, true);
        } else if (!rollActive && lastRollActive) {
            this.rollActiveStates.set(player.name, false);
        }

        if (!isOnGround) {
            this.handleAirState(player, rollState, currentTick, distanceToSolid, isJumping, isGliding, isSneaking, isSprinting, location.y);
        } else if (rollState.rollInputReceived) {
            this.handleLanding(player, rollState, currentTick, dimension, location, solidBlockBelow, isInLiquid);
            this.rollActiveStates.set(player.name, false);
        } else if (!rollState.isRolling) {
            this.resetFallState(rollState);
        }

        if (currentTick % 100 === 0) this.clearBlockCache();
    }

    // Air state handling methods
    handleAirState(player, rollState, currentTick, distanceToSolid, isJumping, isGliding, isSneaking, isSprinting, currentY) {
        if (!rollState.isInAir) {
            rollState.isInAir = true;
            rollState.fallStartY = currentY;
            rollState.maxFallY = currentY;
            rollState.rollInputReceived = false;
            rollState.wasGliding = isGliding;
            rollState.wasSliding = isSneaking && isSprinting;
        }

        if (currentY > rollState.maxFallY) {
            rollState.maxFallY = currentY;
        }

        const fallHeight = rollState.maxFallY - currentY;
        const isAfterDropKick = PlayerStateManager.getState(player.name)?.dropKick?.isInDropKick || false;
        const validFallHeight = isAfterDropKick ? fallHeight >= 2 : fallHeight >= 3;

        if (!rollState.isRolling && !rollState.rollInputReceived) {
            if (isJumping && validFallHeight && distanceToSolid !== Infinity && distanceToSolid >= 0 && distanceToSolid <= 2) {
                rollState.rollInputReceived = true;
                rollState.fallStartY = currentY;
                rollState.rollInputTime = currentTick;
                this.manageTimer(player, this.fallDamageTimers, () => {}, 0);
                player.dimension.runCommand('gamerule falldamage false');
            } else if ((isGliding || rollState.wasGliding) && isJumping && distanceToSolid !== Infinity && distanceToSolid <= 2) {
                rollState.rollInputReceived = true;
                rollState.fallStartY = currentY;
                rollState.rollInputTime = currentTick;
                this.manageTimer(player, this.fallDamageTimers, () => {}, 0);
                player.dimension.runCommand('gamerule falldamage false');
            }
        }

        if (isGliding) rollState.wasGliding = true;
        if (isSneaking && isSprinting) rollState.wasSliding = true;
        else if (player.isOnGround) rollState.wasSliding = false;
    }

    // Start methods
    startWithAnimation(player, currentTick, dimension, location) {
        const rollState = this.validateState(player);
    
        const scan = blockScanner.getPlayerScan(player, currentTick);
        if (!rollState || scan.isInLiquid) return;

        this.resetFallState(rollState);
        rollState.isRolling = true;
        rollState.rollStartTime = currentTick;
        rollState.lastUpdateTime = currentTick;

        this.manageTimer(player, this.rollInputTimers, () => {}, 0);
        system.run(() => {
            player.playAnimation("animation.humanoid.roll", {
                blendOutTime: 0.2
            });
        });

        this.manageTimer(player, this.fallDamageTimers, () => {}, 0);

        const inputPerms = player.inputPermissions;
        inputPerms.jumpEnabled = false;
        dimension.spawnParticle("move:drop", location);
        player.playSound("damage.fallsmall", { volume: 0.6, pitch: 0.8 });

        PlayerStateManager.setRollState(player.name, currentTick);
        this.startContinuousKnockback(player);

        this.manageTimer(player, this.fallDamageTimers, () => {
            this.endRoll(player, system.currentTick);
        }, CONFIG.ROLL.DURATION);
    }

    // Landing methods
    handleLanding(player, rollState, currentTick, dimension, location, solidBlockBelow, isInLiquid) {
        if (isInLiquid) {
            player.dimension.runCommand('gamerule falldamage true');
            this.resetFallState(rollState);
            return;
        }

        if (!rollState.isRolling && rollState.rollInputReceived && solidBlockBelow) {
            this.startWithAnimation(player, currentTick, dimension, location);
            rollState.rollInputReceived = false;
        }
    }

    // Knockback methods
    startContinuousKnockback(player) {
        const rollState = this.validateState(player);
        if (!rollState) return;
        
        if (this.knockbackIntervals.has(player.name)) {
            system.clearRun(this.knockbackIntervals.get(player.name));
            this.knockbackIntervals.delete(player.name);
        }
        
        const power = 0.8;
        const reduction = power * (3 / 100);
        const totalTicks = CONFIG.ROLL.DURATION;
        let tickCount = 0;
        
        const intervalId = system.runInterval(() => {
            if (++tickCount >= totalTicks || power - reduction * tickCount <= 0 || !rollState.isRolling) {
                system.clearRun(intervalId);
                this.knockbackIntervals.delete(player.name);
                return;
            }
            
            const { y: yaw } = player.getRotation();
            const yawRad = (yaw + 90) * Math.PI / 180;
            const pos = player.location;
            const adjustedPower = power * (1 - tickCount / totalTicks);
            
            const scan = blockScanner.getPlayerScan(player, system.currentTick);
            const hasObstacle = scan.hasObstacle;
            const isOneBlockGap = scan.canRoll;
            const isNotInGap = !blockScanner.checkRollGap(player, false);
            
            if (isOneBlockGap && isNotInGap && !hasObstacle) {
                this.tryTeleport(player, system.currentTick, pos, yawRad);
            }
            
            player.applyKnockback(
                Math.cos(yawRad) * adjustedPower,
                Math.sin(yawRad) * adjustedPower,
                adjustedPower,
                -0.12
            );
        });
        
        this.knockbackIntervals.set(player.name, intervalId);
    }

    // End methods
    endRoll(player, currentTick) {
        const rollState = this.validateState(player);
        if (!rollState) return;

        this.manageTimer(player, this.effectTimers, () => {}, 0);
        this.manageTimer(player, this.fallDamageTimers, () => {}, 0);
        this.manageTimer(player, this.knockbackIntervals, () => {}, 0);

        rollState.isRolling = false;
        rollState.lastRollTime = currentTick;
        const inputPerms = player.inputPermissions;
        inputPerms.jumpEnabled = true;
        this.resetFallState(rollState);
        PlayerStateManager.clearRollState(player.name, currentTick);

        system.run(() => {
            player.playAnimation("animation.humanoid.roll_end", {
                blendOutTime: 0.3
            });
        });

        if (!rollState.rollInputReceived) {
            this.manageTimer(player, this.fallDamageTimers, () => {
                const state = PlayerStateManager.getState(player.name);
                if (!state?.roll?.isRolling && !state?.roll?.rollInputReceived) {
                    player.dimension.runCommand('gamerule falldamage true');
                }
            }, 10);
        }
    }

    // Utility methods
    resetFallState(rollState) {
        rollState.isInAir = false;
        rollState.validFallHeight = false;
        rollState.rollInputReceived = false;
        rollState.fallStartY = 0;
        rollState.maxFallY = 0;
        rollState.rollInputTime = undefined;
        rollState.wasGliding = false;
        rollState.wasSliding = false;
    }

    tryTeleport(player, currentTick, location, yawRad) {
        const setLocation = {
            x: location.x + Math.cos(yawRad) * 0.1,
            y: location.y,
            z: location.z + Math.sin(yawRad) * 0.1
        };
        if (player.tryTeleport(setLocation)) {
            player.teleport(setLocation);
            return true;
        }
        return false;
    }

    manageTimer(player, timerMap, callback, delay) {
        if (timerMap.has(player.name)) {
            system.clearRun(timerMap.get(player.name));
            timerMap.delete(player.name);
        }
        const timerId = system.runTimeout(() => {
            callback();
            timerMap.delete(player.name);
        }, delay);
        timerMap.set(player.name, timerId);
    }

    clearBlockCache() {
        blockScanner.clearCache();
    }
}

export default RollHandler;