import { MolangVariableMap, world, system } from '@minecraft/server';
import CONFIG from '../config.js';
import { BLOCK_SOUND_MAPPING, SOUND_CONFIGS, DEFAULT_SOUND } from '../utils/Data.js';
import blockScanner from '../utils/BlockScanner.js';
import { getCachedBlock } from '../utils/blockCache.js';
import PlayerStateManager from '../managers/PlayerStateManager.js';

class SlideHandler {
    constructor() {
        this.lastTick = new Map();
        this.effectTimers = new Map();
    }

    validateState(player) {
        return PlayerStateManager.getState(player.name)?.slide ?? null;
    }

    handleSlide(player, currentTick) {
        const lastTick = this.lastTick.get(player.name) ?? -1;
        if (currentTick === lastTick) return;
        this.lastTick.set(player.name, currentTick);

        const slide = this.validateState(player);
        if (!slide) return;

        const { dimension, location } = player;
        const state = PlayerStateManager.getState(player.name);
        const dropKick = state.dropKick || {};

        const scan = blockScanner.getPlayerScan(player, currentTick);
        const { ground, isInLiquid, hasObstacle, blockGap, canSlide } = scan;
        const { distance } = ground;
        const isOneBlockGap = blockGap.exists;

        if (dropKick.isInDropKick) {
            if (slide.isSliding) this.endSlide(player, currentTick);
            slide.slideBlockedDuringDropKick = true;
            return;
        }

        if (slide.lastSlideTime && currentTick - slide.lastSlideTime < CONFIG.SLIDE.COOLDOWN) return;

        const isSprinting = player.isSprinting;
        const isSneaking = player.isSneaking;
        const isOnGround = player.isOnGround;

        if (isSprinting) {
            slide.wasSprintingFirst = true;
            slide.lastSprintTime = currentTick;
        }

        const sprintBuffer = 4;
        const wasRecentlySprinting = currentTick - slide.lastSprintTime < sprintBuffer;
        if (!wasRecentlySprinting && !isSneaking) slide.wasSprintingFirst = false;

        if (wasRecentlySprinting && isSneaking && slide.wasSprintingFirst && !slide.isSliding && isOnGround && !isInLiquid && ground.block && !hasObstacle) {
            this.startWithAnimation(player, currentTick, dimension, location);
        }

        if (slide.isSliding) {
            this.updateSlide(player, currentTick, state, dimension, location, isInLiquid, isOneBlockGap, distance, hasObstacle);
        }

        if (currentTick % 100 === 0) this.clearCaches();
    }

    startWithAnimation(player, currentTick, dimension, location) {
        const state = PlayerStateManager.getState(player.name);
        
        const scan = blockScanner.getPlayerScan(player, currentTick);
        if (!state || state.roll?.isRolling || scan.isInLiquid) return;

        state.slide.isSliding = true;
        state.slide.startTime = currentTick;
        system.run(() => {
            player.playAnimation("animation.humanoid.slide", { 
                blendOutTime: 0.3 
            });
        });

        PlayerStateManager.setSlideState(player.name, {
            startTime: currentTick,
            slideDuration: 14,
            currentKnockback: 0.85,
            lastUpdateTime: currentTick,
        });
    }

    updateSlide(player, currentTick, state, dimension, location, isInLiquid, isOneBlockGap, distance, hasObstacle) {
        const slideInfo = state.slide.slideInfo;
        if (!slideInfo) return;

        if (isInLiquid || currentTick - slideInfo.startTime > 14 || !player.isSneaking || hasObstacle) {
            this.endSlide(player, currentTick);
            return;
        }

        const timeSinceLastUpdate = (currentTick - slideInfo.lastUpdateTime) / 20;
        slideInfo.currentKnockback *= Math.pow(1 - (25 / 100), timeSinceLastUpdate);
        slideInfo.lastUpdateTime = currentTick;

        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;
        const elevation = distance === Infinity || distance > 0.1 ? -0.14 : 0.04;

        const knockbackX = Math.cos(yawRad) * slideInfo.currentKnockback;
        const knockbackZ = Math.sin(yawRad) * slideInfo.currentKnockback;
        slideInfo.lastKnockbackTime = slideInfo.lastKnockbackTime || currentTick;
        player.applyKnockback(knockbackX, knockbackZ, slideInfo.currentKnockback, elevation);
        slideInfo.lastKnockbackTime = currentTick;

        const blockBelow = getCachedBlock(dimension, { 
            x: Math.floor(location.x), 
            y: Math.floor(location.y) - 1, 
            z: Math.floor(location.z) 
        });
        const blockSound = BLOCK_SOUND_MAPPING[blockBelow?.typeId] || DEFAULT_SOUND.sound;
        const soundConfig = SOUND_CONFIGS[blockSound] || DEFAULT_SOUND;
        dimension.playSound(blockSound, location, { 
            pitch: soundConfig.pitch, 
            volume: soundConfig.volume 
        });

        if (isOneBlockGap && !blockScanner.checkBlockGap(player, false) && !hasObstacle) {
            this.tryTeleport(player, currentTick, location, yawRad);
        }

        this.spawnParticle(player);
    }

    endSlide(player, currentTick) {
        const state = PlayerStateManager.getState(player.name);
        if (!state) return;

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }

        state.slide.isSliding = false;
        state.slide.lastSlideTime = currentTick;
        PlayerStateManager.clearSlideState(player.name, currentTick);

        system.run(() => {
            player.playAnimation("animation.humanoid.slide.end", { 
                blendOutTime: 0.3 
            });
        });
    }

    tryTeleport(player, currentTick, location, yawRad) {
        const setLocation = {
            x: location.x + Math.cos(yawRad) * 0.1,
            y: location.y,
            z: location.z + Math.sin(yawRad) * 0.1
        };
        if (player.tryTeleport(setLocation)) {
            player.teleport(setLocation);
            return true;
        }
        return false;
    }

    spawnParticle(player) {
        const playerVelocity = player.getVelocity();
        const velocity = Math.sqrt(playerVelocity.x ** 2 + playerVelocity.z ** 2);

        if (velocity > 0.15) {
            const { location, dimension } = player;
            const { y: yaw } = player.getRotation();
            const rotationRad = (yaw + 90) * Math.PI / 180;
            const dirX = Math.cos(rotationRad);
            const dirZ = Math.sin(rotationRad);
            const molang = new MolangVariableMap();
            molang.setFloat("variable.direction_x", dirX);
            molang.setFloat("variable.direction_y", 0);
            molang.setFloat("variable.direction_z", dirZ);
            const spawnPos = {
                x: location.x - (playerVelocity.x * 0.3),
                y: location.y,
                z: location.z - (playerVelocity.z * 0.3),
            };
            dimension.spawnParticle("move:trail", spawnPos, molang);
        }
    }

    clearCaches() {
        blockScanner.clearCache();
    }
}

export default SlideHandler;