import { world, system } from '@minecraft/server';
import blockScanner from '../utils/BlockScanner.js';
import PlayerStateManager from '../managers/PlayerStateManager.js';
import CONFIG from '../config.js';

class VaultHandler {
    constructor() {
        this.fallDamageTimers = new Map();
        this.lastTick = new Map();
        this.effectTimers = new Map();
    }

    // Validation methods
    validateState(player) {
        return PlayerStateManager.getState(player.name)?.vault ?? null;
    }

    // Main handler method
    handleVault(player, currentTick) {
        const lastTick = this.lastTick.get(player.name) ?? -1;
        if (currentTick === lastTick) return;
        this.lastTick.set(player.name, currentTick);

        const vaultState = this.validateState(player);
        if (!vaultState) return;

        const { dimension, location } = player;
        const state = PlayerStateManager.getState(player.name);

        // ONE scan call - gets cached comprehensive data
        const scan = blockScanner.getPlayerScan(player, currentTick);
        const { isInLiquid, ceiling, vaultCheck } = scan;
        
        const isSprinting = player.isSprinting;
        const isJumping = player.isJumping;
        const isOnCooldown = vaultState.lastVaultTime && (currentTick - vaultState.lastVaultTime) < CONFIG.VAULT.COOLDOWN;
        const isBlocked = state.roll?.isRolling || state.roll?.rollInputReceived || 
                         state.slide?.isSliding || state.dropKick?.isInDropKick;

        if (vaultState.isVaulting) {
            if (currentTick - vaultState.vaultStartTime > CONFIG.VAULT.DURATION) {
                this.endVault(player, currentTick);
            }
            return;
        }

        if (isOnCooldown) {
            vaultState.cooldownNotified = true;
            return;
        }
        vaultState.cooldownNotified = false;

        if (!isSprinting || !isJumping || isBlocked || isInLiquid) {
            vaultState.preventionNotified = isBlocked;
            return;
        }

        this.startWithAnimation(player, currentTick, state, dimension, location);

        if (currentTick % 100 === 0) this.clearCaches();
    }

    // Start methods
    startWithAnimation(player, currentTick, state, dimension, location) {
        const vaultState = this.validateState(player);
        if (!vaultState || state.roll?.rollInputReceived || state.roll?.isRolling) return;

        // Use comprehensive scan data
        const scan = blockScanner.getPlayerScan(player, currentTick);
        const { valid, tooHigh, tooLow } = scan.vaultCheck;
        const isBlockAboveHead = scan.ceiling.hasBlock;
        
        if (!valid || isBlockAboveHead) {
            vaultState.tooHighNotified = tooHigh || isBlockAboveHead || false;
            return;
        }
        vaultState.tooHighNotified = false;

        Object.assign(vaultState, {
            isVaulting: true,
            vaultStartTime: currentTick,
            lastVaultTime: currentTick
        });

        PlayerStateManager.setVaultState(player.name, null);

        if (state.roll) {
            this.resetFallState(state.roll);
            player.dimension.runCommand('gamerule falldamage false');
        }

        const boostStrength = 2;
        const upwardBoost = 0.52;
        player.playSound("game.player.attack.strong", { pitch: 1.5, volume: 0.6 });
        system.run(() => {
            player.playAnimation("animation.humanoid.vault", {
                blendOutTime: 0.2
            });
        });

        player.applyKnockback(
            player.getViewDirection().x * boostStrength,
            player.getViewDirection().z * boostStrength,
            0.2,
            upwardBoost
        );
    }

    // End methods
    endVault(player, currentTick) {
        const state = PlayerStateManager.getState(player.name);
        const vaultState = this.validateState(player);
        if (!state || !vaultState) return;

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        if (this.fallDamageTimers.has(player.name)) {
            system.clearRun(this.fallDamageTimers.get(player.name));
            this.fallDamageTimers.delete(player.name);
        }

        vaultState.isVaulting = false;
        PlayerStateManager.clearVaultState(player.name, currentTick);

        if (state.roll && !state.roll.isRolling && !state.roll.rollInputReceived) {
            const timerId = system.runTimeout(() => {
                if (!state.roll?.isRolling && !state.roll?.rollInputReceived) {
                    player.dimension.runCommand('gamerule falldamage true');
                }
                this.fallDamageTimers.delete(player.name);
            }, 10);
            this.fallDamageTimers.set(player.name, timerId);
        }
    }

    // Utility methods
    resetFallState(rollState) {
        Object.assign(rollState, {
            isInAir: false,
            validFallHeight: false,
            rollInputReceived: false,
            fallStartY: 0
        });
    }

    clearCaches() {
        blockScanner.clearCache();
    }
}

export default VaultHandler;