import { world, system } from '@minecraft/server';
import blockScanner from '../utils/BlockScanner.js';
import { BLOCK_SOUND_MAPPING, SOUND_CONFIGS, DEFAULT_SOUND } from '../utils/Data.js';
import PlayerStateManager from '../managers/PlayerStateManager.js';
import CONFIG from '../config.js';

class WallRunHandler {
    constructor() {
        this.fallDamageTimers = new Map();
        this.lastTick = new Map();
        this.effectTimers = new Map();
        this.wallContactTimers = new Map();
        this.mustTouchGroundPlayers = new Set();
        this.lastJumpState = new Map();

        this.WALL_ADHESION = 0.25;
        this.MAX_DISTANCE = 0.4;
        this.GRACE_PERIOD = 5;
        this.WALL_JUMP_FORCE = 0.8;
        this.WALL_JUMP_UPWARD = 0.4;
        this.WALL_JUMP_FORWARD = 0.4;
        this.MIN_OPPOSITE_WALL_DISTANCE = 1;
        this.FALL_DAMAGE_DELAY = 10;
        this.WALL_JUMP_COOLDOWN = 2; // 2-tick cooldown after wall jump
    }

    validateState(player) {
        return PlayerStateManager.getState(player.name)?.wallrun ?? null;
    }

    handleWallRun(player, currentTick) {
        const lastTick = this.lastTick.get(player.name) ?? -1;
        if (currentTick === lastTick) return;
        this.lastTick.set(player.name, currentTick);

        const wallrun = this.validateState(player);
        if (!wallrun) return;

        const { dimension, location } = player;
        const state = PlayerStateManager.getState(player.name);

        const scan = blockScanner.getPlayerScan(player, currentTick);
        const { isInLiquid, heightAboveGround, wallSide } = scan;

        const isSprinting = player.isSprinting;
        const isOnGround = player.isOnGround;
        const isGliding = player.isGliding;
        const isJumping = player.isJumping;
        const mustTouchGround = this.mustTouchGroundPlayers.has(player.name);
        const isBlocked = state.vault?.isVaulting || state.dropKick?.isInDropKick ||
                         state.slide?.isSliding || state.roll?.isRolling;
        const isOnWallJumpCooldown = wallrun.lastWallJumpTime && 
                                    (currentTick - wallrun.lastWallJumpTime) < this.WALL_JUMP_COOLDOWN;

        if (mustTouchGround) {
            if (isOnGround) {
                this.mustTouchGroundPlayers.delete(player.name);
            } else {
                return;
            }
        }

        if (wallrun.isWallRunning) {
            this.handleWallJumpInput(player, currentTick, state, dimension, location);
            this.updateWallRun(player, currentTick, state, dimension, location);
            return;
        }

        if (isBlocked || !isSprinting || isOnGround ||
            isGliding || heightAboveGround < 2 || isInLiquid || isOnWallJumpCooldown) return;

        this.startWithAnimation(player, currentTick, state, dimension, location);

        if (currentTick % 100 === 0) this.clearCaches();
    }

    checkOppositeWallDistance(player, currentWallSide) {
        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;

        const oppositeSideDirX = currentWallSide === "left" ?
            Math.cos(yawRad + Math.PI / 2) :
            Math.cos(yawRad - Math.PI / 2);
        const oppositeSideDirZ = currentWallSide === "left" ?
            Math.sin(yawRad + Math.PI / 2) :
            Math.sin(yawRad - Math.PI / 2);

        const rayStart = {
            x: player.location.x,
            y: player.location.y + 1.0,
            z: player.location.z
        };

        const rayResult = player.dimension.getBlockFromRay(
            rayStart,
            { x: oppositeSideDirX, y: 0, z: oppositeSideDirZ },
            { maxDistance: this.MIN_OPPOSITE_WALL_DISTANCE, includePassableBlocks: false, includeLiquidBlocks: false }
        );

        return !rayResult?.block || rayResult.block.isAir || rayResult.block.isLiquid;
    }

    performWallJump(player, currentTick, wallrun, dimension, location) {
        if (!this.checkOppositeWallDistance(player, wallrun.side)) {
            player.playSound("game.player.attack.nodamage", { volume: 0.5, pitch: 0.8 });
            return;
        }

        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;

        const oppositeSideDirX = wallrun.side === "left" ?
            Math.cos(yawRad + Math.PI / 2) :
            Math.cos(yawRad - Math.PI / 2);
        const oppositeSideDirZ = wallrun.side === "left" ?
            Math.sin(yawRad + Math.PI / 2) :
            Math.sin(yawRad - Math.PI / 2);

        const forwardX = Math.cos(yawRad) * this.WALL_JUMP_FORWARD;
        const forwardZ = Math.sin(yawRad) * this.WALL_JUMP_FORWARD;
        const lateralX = oppositeSideDirX * this.WALL_JUMP_FORCE;
        const lateralZ = oppositeSideDirZ * this.WALL_JUMP_FORCE;

        player.applyKnockback(
            forwardX + lateralX,
            forwardZ + lateralZ,
            this.WALL_JUMP_FORCE,
            this.WALL_JUMP_UPWARD
        );

        player.playSound("game.player.attack.strong", { volume: 0.8, pitch: 1.2 });
        dimension.spawnParticle("parkour:wallrun_dust", {
            x: location.x + oppositeSideDirX * 0.3,
            y: location.y + 1.1,
            z: location.z + oppositeSideDirZ * 0.3
        });

        this.mustTouchGroundPlayers.delete(player.name);
        wallrun.lastWallJumpTime = currentTick; // Set cooldown timestamp
        this.endWallRun(player, currentTick, true);
    }

    handleWallJumpInput(player, currentTick, state, dimension, location) {
        const wallrun = this.validateState(player);
        if (!wallrun || !wallrun.wallRunInfo) return;

        const isJumping = player.isJumping;
        const lastJumpState = this.lastJumpState.get(player.name) ?? false;

        if (isJumping && !lastJumpState) {
            this.performWallJump(player, currentTick, wallrun, dimension, location);
        }

        this.lastJumpState.set(player.name, isJumping);
    }

    startWithAnimation(player, currentTick, state, dimension, location) {
        const wallrun = this.validateState(player);
        if (!wallrun || state.roll?.rollInputReceived || state.roll?.isRolling) return;

        const scan = blockScanner.getPlayerScan(player, currentTick);
        const side = scan.wallSide;

        if (!side) return;

        this.clearFallDamageTimer(player);
        player.dimension.runCommand('gamerule falldamage false');

        const knockbackPower = 0.6;
        PlayerStateManager.setWallRunState(player.name, side, {
            startTime: currentTick,
            currentKnockback: knockbackPower,
            lastUpdateTime: currentTick,
            lastSoundTime: 0,
            lastParticleTime: 0,
            lastFootParticleTime: 0,
            wallContactLost: false,
            lastWallCheck: currentTick
        });

        wallrun.isWallRunning = true;

        const animation = side === "left" ? "animation.humanoid.wallrun.left" : "animation.humanoid.wallrun.right";
        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }

        const viewDir = player.getViewDirection();
        player.applyKnockback(
            viewDir.x * knockbackPower,
            viewDir.z * knockbackPower,
            knockbackPower,
            0
        );

        system.run(() => {
            if (this.validateState(player)?.isWallRunning) {
                player.playAnimation(animation, {
                    blendOutTime: 0.3
                });
                player.applyKnockback(
                    viewDir.x * knockbackPower * 0.8,
                    viewDir.z * knockbackPower * 0.8,
                    knockbackPower * 0.8,
                    0
                );
            }
        });

        this.lastJumpState.set(player.name, player.isJumping);
    }

    applyWallRunForces(player, wallrun, wallRunInfo, currentTick) {
        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;
        const sideDirX = wallrun.side === "left" ?
            Math.cos(yawRad - Math.PI / 2) :
            Math.cos(yawRad + Math.PI / 2);
        const sideDirZ = wallrun.side === "left" ?
            Math.sin(yawRad - Math.PI / 2) :
            Math.sin(yawRad + Math.PI / 2);

        const timeSinceLastUpdate = (currentTick - wallRunInfo.lastUpdateTime) / 20;
        const decayRate = 1 - (10 / 100);
        wallRunInfo.currentKnockback *= Math.pow(decayRate, timeSinceLastUpdate);
        wallRunInfo.lastUpdateTime = currentTick;

        const forwardX = Math.cos(yawRad) * wallRunInfo.currentKnockback;
        const forwardZ = Math.sin(yawRad) * wallRunInfo.currentKnockback;
        const gravity = -0.03;
        const lateralDriftX = sideDirX * this.WALL_ADHESION;
        const lateralDriftZ = sideDirZ * this.WALL_ADHESION;

        player.applyKnockback(
            forwardX + lateralDriftX,
            forwardZ + lateralDriftZ,
            wallRunInfo.currentKnockback,
            gravity
        );

        const scan = blockScanner.getPlayerScan(player, currentTick);
        const currentWallSide = scan.wallSide;
        if (currentWallSide === wallrun.side) {
            wallRunInfo.wallContactLost = false;
            wallRunInfo.lastWallCheck = currentTick;
        } else if (!wallRunInfo.wallContactLost) {
            wallRunInfo.wallContactLost = true;

            player.applyKnockback(
                lateralDriftX * 1.5,
                lateralDriftZ * 1.5,
                0.3,
                0
            );

            if (this.wallContactTimers.has(player.name)) {
                system.clearRun(this.wallContactTimers.get(player.name));
            }

            const timerId = system.runTimeout(() => {
                if (wallRunInfo.wallContactLost &&
                    currentTick - wallRunInfo.lastWallCheck > this.GRACE_PERIOD) {
                    this.endWallRun(player, currentTick);
                }
                this.wallContactTimers.delete(player.name);
            }, this.GRACE_PERIOD);

            this.wallContactTimers.set(player.name, timerId);
        }
    }

    spawnParticlesAndPlaySounds(player, wallrun, wallRunInfo, currentTick, dimension, location) {
        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;
        const sideDirX = wallrun.side === "left" ?
            Math.cos(yawRad - Math.PI / 2) :
            Math.cos(yawRad + Math.PI / 2);
        const sideDirZ = wallrun.side === "left" ?
            Math.sin(yawRad - Math.PI / 2) :
            Math.sin(yawRad + Math.PI / 2);

        const handOffsetY = 1.1;
        const handOffsetDistance = 0.3;
        const handParticlePos = {
            x: location.x + sideDirX * handOffsetDistance,
            y: location.y + handOffsetY,
            z: location.z + sideDirZ * handOffsetDistance
        };
        if (currentTick - (wallRunInfo.lastParticleTime || 0) >= 1) {
            dimension.spawnParticle("parkour:wallrun_dust", handParticlePos);
            wallRunInfo.lastParticleTime = currentTick;
        }

        const footOffsetY = 0.1;
        const footOffsetDistance = 0.25;
        const footParticlePos = {
            x: location.x + sideDirX * footOffsetDistance,
            y: location.y + footOffsetY,
            z: location.z + sideDirZ * handOffsetDistance
        };
        if (currentTick - (wallRunInfo.lastFootParticleTime || 0) >= 2) {
            dimension.spawnParticle("parkour:wallrun_foot_dust", footParticlePos);
            wallRunInfo.lastFootParticleTime = currentTick;
        }

        if (currentTick - wallRunInfo.lastSoundTime >= 3) {
            const scan = blockScanner.getPlayerScan(player, currentTick);
            const wallBlock = scan.walls[wallrun.side]?.block;
            if (wallBlock) {
                const blockSound = BLOCK_SOUND_MAPPING[wallBlock.typeId] || DEFAULT_SOUND.sound;
                const soundConfig = SOUND_CONFIGS[blockSound] || DEFAULT_SOUND;
                dimension.playSound(blockSound, location, {
                    pitch: soundConfig.pitch,
                    volume: soundConfig.volume
                });
                wallRunInfo.lastSoundTime = currentTick;
            }
        }
    }

    updateWallRun(player, currentTick, state, dimension, location) {
        const wallrun = this.validateState(player);
        if (!wallrun || !wallrun.wallRunInfo) return;
        const wallRunInfo = wallrun.wallRunInfo;

        if (state.vault?.isVaulting) {
            this.endWallRun(player, currentTick);
            return;
        }

        const scan = blockScanner.getPlayerScan(player, currentTick);
        const currentWallSide = scan.wallSide;

        if (player.isOnGround ||
            (currentTick - wallRunInfo.startTime) > CONFIG.WALLRUN.DURATION ||
            !player.isSprinting ||
            player.isGliding ||
            scan.isInLiquid ||
            currentWallSide !== wallrun.side ||
            scan.heightAboveGround < 1) {
            this.endWallRun(player, currentTick);
            return;
        }

        this.applyWallRunForces(player, wallrun, wallRunInfo, currentTick);
        this.spawnParticlesAndPlaySounds(player, wallrun, wallRunInfo, currentTick, dimension, location);
    }

    clearFallDamageTimer(player) {
        if (this.fallDamageTimers.has(player.name)) {
            system.clearRun(this.fallDamageTimers.get(player.name));
            this.fallDamageTimers.delete(player.name);
        }
    }

    startFallDamageTimer(player, state) {
        this.clearFallDamageTimer(player);

        const timerId = system.runTimeout(() => {
            if (!state.roll?.isRolling && !state.roll?.rollInputReceived && !this.validateState(player)?.isWallRunning) {
                player.dimension.runCommand('gamerule falldamage true');
            }
            this.fallDamageTimers.delete(player.name);
        }, this.FALL_DAMAGE_DELAY);

        this.fallDamageTimers.set(player.name, timerId);
    }

    endWallRun(player, currentTick, isWallJump = false) {
        const state = PlayerStateManager.getState(player.name);
        if (!state) return;

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }

        if (this.wallContactTimers.has(player.name)) {
            system.clearRun(this.wallContactTimers.get(player.name));
            this.wallContactTimers.delete(player.name);
        }

        const wallrun = this.validateState(player);
        if (wallrun && wallrun.isWallRunning) {
            const endAnimation = wallrun.side === "left" ?
                "animation.humanoid.wallrun.left.end" :
                "animation.humanoid.wallrun.right.end";
            player.playAnimation(endAnimation, { blendOutTime: 0.3 });
        }

        this.lastJumpState.delete(player.name);

        if (!isWallJump) {
            this.mustTouchGroundPlayers.add(player.name);
        }

        wallrun.isWallRunning = false;
        if (state.roll) this.resetFallState(state.roll);
        PlayerStateManager.clearWallRunState(player.name, currentTick);

        if (state.roll && !state.roll.isRolling && !state.roll.rollInputReceived) {
            this.startFallDamageTimer(player, state);
        } else {
            this.startFallDamageTimer(player, state);
        }
    }

    resetFallState(rollState) {
        Object.assign(rollState, {
            isInAir: false,
            validFallHeight: false,
            rollInputReceived: false,
            fallStartY: 0
        });
    }

    clearCaches() {
        blockScanner.clearCache();
    }
}

export default WallRunHandler;