import { world, system } from '@minecraft/server';
import Logger from '../utils/Logger.js';

class PlayerStateManager {
    constructor() {
        this.states = new Map();
        this.initializeEventListeners();
    }

    initializeEventListeners() {
        world.afterEvents.playerSpawn.subscribe((event) => {
            const playerName = event.player.name;
            if (!this.states.has(playerName)) {
                this.initializePlayer(playerName);
            }
        });

        world.afterEvents.playerLeave.subscribe((event) => {
            this.clearPlayerStates(event.playerName);
        });
    }

    getState(playerName) {
        if (!this.states.has(playerName)) {
            this.initializePlayer(playerName);
        }
        return this.states.get(playerName);
    }

    initializePlayer(playerName) {
        const initialState = {
            slide: {
                wasSprintingFirst: false,
                lastSprintTime: 0,
                isSliding: false,
                lastSlideTime: null,
                lastCooldownMessage: 0,
                startTime: 0,
                slideInfo: null
            },
            roll: {
                isInAir: false,
                isFalling: false,
                fallStartY: 0,
                rollInputReceived: false,
                isRolling: false,
                rollStartTime: 0,
                lastY: 0,
                validFallHeight: false,
                lastRollTime: null,
                lastCooldownMessage: 0,
                lastUpdateTime: 0,
                inLiquid: false,
                wasGliding: false,
                rollInputTime: null
            },
            vault: {
                isVaulting: false,
                wasVaulting: false,
                mustTouchGround: false,
                lastVaultTime: null,
                lastCooldownMessage: 0,
                lastUpdateTime: 0,
                previousPreventState: false,
                cooldownNotified: false,
                vaultInfo: null
            },
            wallrun: {
                isWallRunning: false,
                side: null,             
                startTime: 0,
                lastWallRunTime: null,
                lastCooldownMessage: 0,
                currentKnockback: 0,
                wallRunInfo: null
            },
            crawl: {
                isCrawling: false,
                startTime: 0,
                lastCrawlTime: null,
                lastCooldownMessage: 0,
                crawlInfo: null
            }
        };

        this.states.set(playerName, initialState);
        Logger.debug(`Initialized state for player ${playerName}`, "SYSTEM");
    }

    setSlideState(playerName, slideInfo) {
        const state = this.getState(playerName);
        state.slide.isSliding = true;
        state.slide.startTime = system.currentTick;
        state.slide.slideInfo = slideInfo;
        Logger.debug(`Player ${playerName} started sliding`, "SLIDE");
    }

    clearSlideState(playerName, currentTick) {
        const state = this.getState(playerName);
        state.slide.isSliding = false;
        state.slide.lastSlideTime = currentTick;
        state.slide.slideInfo = null;
        Logger.debug(`Player ${playerName} slide ended`, "SLIDE");
    }

    setRollState(playerName, currentTick) {
        const state = this.getState(playerName);
        state.roll.isRolling = true;
        state.roll.rollStartTime = currentTick;
        Logger.debug(`Player ${playerName} started rolling`, "ROLL");
    }

    clearRollState(playerName, currentTick) {
        const state = this.getState(playerName);
        state.roll.isRolling = false;
        state.roll.lastRollTime = currentTick;
        Logger.debug(`Player ${playerName} roll ended`, "ROLL");
    }

    setVaultState(playerName, vaultInfo) {
        const state = this.getState(playerName);
        state.vault.isVaulting = true;
        state.vault.vaultInfo = vaultInfo;
        Logger.debug(`Player ${playerName} started vaulting`, "VAULT");
    }

    clearVaultState(playerName, currentTick) {
        const state = this.getState(playerName);
        state.vault.isVaulting = false;
        state.vault.lastVaultTime = currentTick;
        state.vault.vaultInfo = null;
        Logger.debug(`Player ${playerName} vault ended`, "VAULT");
    }

    setWallRunState(playerName, side, wallRunInfo) {
        const state = this.getState(playerName);
        state.wallrun.isWallRunning = true;
        state.wallrun.side = side;
        state.wallrun.startTime = system.currentTick;
        state.wallrun.wallRunInfo = wallRunInfo;
        Logger.debug(`Player ${playerName} started wall running on ${side}`, "WALLRUN");
    }

    clearWallRunState(playerName, currentTick) {
        const state = this.getState(playerName);
        state.wallrun.isWallRunning = false;
        state.wallrun.side = null;
        state.wallrun.startTime = 0;
        state.wallrun.lastWallRunTime = currentTick;
        state.wallrun.wallRunInfo = null;
        Logger.debug(`Player ${playerName} wall run ended`, "WALLRUN");
    }

    setCrawlState(playerName, crawlInfo) {
        const state = this.getState(playerName);
        state.crawl.isCrawling = true;
        state.crawl.startTime = system.currentTick;
        state.crawl.crawlInfo = crawlInfo;
        Logger.debug(`Player ${playerName} started crawling`, "CRAWL");
    }

    clearCrawlState(playerName, currentTick) {
        const state = this.getState(playerName);
        state.crawl.isCrawling = false;
        state.crawl.lastCrawlTime = currentTick;
        state.crawl.crawlInfo = null;
        Logger.debug(`Player ${playerName} crawl ended`, "CRAWL");
    }

    clearPlayerStates(playerName) {
        if (this.states.has(playerName)) {
            this.states.delete(playerName);
            Logger.info(`Cleared all states for player ${playerName}`, "SYSTEM");
        }
    }
}

export default new PlayerStateManager();