import { system } from '@minecraft/server';
import { getCachedBlock, clearBlockCache } from './blockCache.js';

/**
 * Centralized block scanning utility that provides consistent environment detection
 * for all movement mechanics. Uses caching for performance optimization and provides
 * comprehensive scanning to eliminate duplicate scans per tick.
 */
class BlockScanner {
    constructor() {
        this.scanCache = new Map();
        this.comprehensiveCache = new Map(); // Cache for comprehensive scans
        this.lastComprehensiveScan = new Map(); // Track last comprehensive scan tick
        this.FORWARD_DISTANCE = 0.1;
        this.CACHE_LIFETIME = 1; // How many ticks a cache entry remains valid
    }

    /**
     * Get comprehensive scan data for a player - ONE scan per player per tick with ALL data
     * This is the main method handlers should use to get all scanning data efficiently
     * @param {Player} player - The player to scan
     * @param {number} currentTick - Current game tick
     * @returns {Object} Comprehensive scan results with convenience properties
     */
    getPlayerScan(player, currentTick) {
        const cacheKey = player.name;
        
        // Return cached scan if same tick
        if (this.lastComprehensiveScan.get(cacheKey) === currentTick) {
            return this.comprehensiveCache.get(cacheKey);
        }

        // Perform ONE comprehensive scan with ALL options
        const comprehensiveScan = this.scan(player, {
            checkGround: true,
            checkObstacles: true,
            checkCeiling: true,
            checkWalls: true,
            checkGap: true,
            usePlayerForward: true
        });

        // Add convenience properties for easy handler access
        comprehensiveScan.heightAboveGround = comprehensiveScan.ground.distance;
        comprehensiveScan.wallSide = this.detectWall(player);
        comprehensiveScan.canCrawl = this.checkBlockGap(player, false);
        comprehensiveScan.canRoll = this.checkRollGap(player, true);
        comprehensiveScan.vaultCheck = this.checkVaultBlock(player);

        // Add additional convenience properties for handlers
        comprehensiveScan.isOnGround = comprehensiveScan.heightAboveGround < 0.1;
        comprehensiveScan.hasWall = comprehensiveScan.wallSide !== null;
        comprehensiveScan.canSlide = comprehensiveScan.blockGap.exists && !comprehensiveScan.hasObstacle;

        // Cache the result
        this.comprehensiveCache.set(cacheKey, comprehensiveScan);
        this.lastComprehensiveScan.set(cacheKey, currentTick);
        
        return comprehensiveScan;
    }

    /**
     * Main scanning method that performs various environment checks based on options
     * @param {Player} player - The player to scan around
     * @param {Object} options - Configuration options for the scan
     * @param {boolean} [options.checkGround=true] - Whether to check for ground blocks
     * @param {boolean} [options.checkObstacles=true] - Whether to check for obstacles
     * @param {boolean} [options.checkCeiling=false] - Whether to check for blocks above
     * @param {boolean} [options.checkWalls=false] - Whether to check for walls
     * @param {boolean} [options.checkGap=false] - Whether to check for block gaps (crawl/slide)
     * @param {boolean} [options.usePlayerForward=false] - Whether to offset scanning by player's forward direction
     * @returns {Object} Scan results containing detected blocks and environmental data
     */
    scan(player, options = {}) {
        const currentTick = system.currentTick;
        const { location, dimension } = player;
        const { x, y, z } = location;
        
        // Set default options
        const opts = {
            checkGround: true,
            checkObstacles: true, 
            checkCeiling: false,
            checkWalls: false,
            checkGap: false,
            usePlayerForward: false,
            ...options
        };

        // Generate a unique cache key based on player location and tick
        const cacheKey = `${player.name}:${Math.floor(x)},${Math.floor(y)},${Math.floor(z)}:${JSON.stringify(opts)}`;
        
        // Check if we have a valid cached result
        const cached = this.scanCache.get(cacheKey);
        if (cached && currentTick - cached.timestamp < this.CACHE_LIFETIME) {
            return cached.result;
        }

        // Initialize result object
        const result = {
            ground: {
                block: undefined,
                distance: Infinity
            },
            isInLiquid: false,
            hasObstacle: false,
            walls: {
                left: null,
                right: null
            },
            ceiling: {
                block: null,
                distance: Infinity,
                hasBlock: false
            },
            blockGap: {
                exists: false
            }
        };

        // Check if player is in liquid
        const currentBlock = getCachedBlock(dimension, { 
            x: Math.floor(x), 
            y: Math.floor(y), 
            z: Math.floor(z) 
        });
        result.isInLiquid = currentBlock?.isLiquid || false;

        // Standard offsets for ray casting
        const offsets = [-0.29, 0, 0.29];
        
        // Get player rotation if needed
        let yawRad;
        if (opts.checkObstacles || opts.checkWalls || opts.checkGap || opts.usePlayerForward) {
            const { y: yaw } = player.getRotation();
            yawRad = (yaw + 90) * Math.PI / 180;
        }

        // Scan for ground below player
        if (opts.checkGround) {
            for (const dx of offsets) {
                for (const dz of offsets) {
                    const checkPos = { x: x + dx, y, z: z + dz };
                    const downRay = dimension.getBlockFromRay(
                        checkPos, 
                        { x: 0, y: -1, z: 0 }, 
                        { maxDistance: 2, includePassableBlocks: false, includeLiquidBlocks: false }
                    );
                    
                    if (downRay?.block && !downRay.block.isAir && !downRay.block.isLiquid) {
                        const distance = y - (downRay.block.location.y + downRay.faceLocation.y);
                        if (distance >= 0 && distance < result.ground.distance) {
                            result.ground.distance = distance;
                            result.ground.block = downRay.block;
                        }
                    }
                }
            }
        }

        // Scan for obstacles in front of player
        if (opts.checkObstacles) {
            for (const dx of offsets) {
                for (const dz of offsets) {
                    const obstaclePos = { x: x + dx, y: y + 0.1, z: z + dz };
                    const obstacleRay = dimension.getBlockFromRay(
                        obstaclePos, 
                        { 
                            x: Math.cos(yawRad), 
                            y: 0, 
                            z: Math.sin(yawRad) 
                        }, 
                        { maxDistance: 1.0, includePassableBlocks: false, includeLiquidBlocks: false }
                    );
                    
                    if (obstacleRay && !obstacleRay.block.isAir && !obstacleRay.block.isLiquid) {
                        result.hasObstacle = true;
                        result.obstacle = {
                            block: obstacleRay.block,
                            distance: obstacleRay.distance
                        };
                        break;
                    }
                }
                if (result.hasObstacle) break;
            }
        }

        // Scan for ceiling above player
        if (opts.checkCeiling) {
            for (const dx of offsets) {
                for (const dz of offsets) {
                    const ceilingPos = { x: x + dx, y: y + 1.8, z: z + dz };
                    const upRay = dimension.getBlockFromRay(
                        ceilingPos, 
                        { x: 0, y: 1, z: 0 }, 
                        { maxDistance: 1, includePassableBlocks: false, includeLiquidBlocks: false }
                    );
                    
                    if (upRay?.block && !upRay.block.isAir && !upRay.block.isLiquid) {
                        const distance = (upRay.block.location.y + upRay.faceLocation.y) - (y + 1.8);
                        if (distance >= 0 && distance < result.ceiling.distance) {
                            result.ceiling.distance = distance;
                            result.ceiling.block = upRay.block;
                            result.ceiling.hasBlock = true;
                        }
                    }
                }
            }
        }

        // Check for walls on the sides (for wall running) - Now using body rotation instead of view direction
        if (opts.checkWalls) {
            // Use body rotation (yaw) which we already calculated
            const angles = [-40 * Math.PI / 180, -25 * Math.PI / 180, 0, 25 * Math.PI / 180, 40 * Math.PI / 180]; 
            const directions = [
                { side: "right", baseAngle: yawRad + Math.PI / 2 },
                { side: "left", baseAngle: yawRad - Math.PI / 2 }
            ];

            for (const { side, baseAngle } of directions) {
                for (const angleOffset of angles) {
                    const rayAngle = baseAngle + angleOffset;
                    const rayDirX = Math.cos(rayAngle);
                    const rayDirZ = Math.sin(rayAngle);

                    for (const heightOffset of offsets) {
                        if (heightOffset < 0) continue;
                        const rayStart = { 
                            x: location.x, 
                            y: location.y + heightOffset, 
                            z: location.z 
                        };

                        const rayResult = dimension.getBlockFromRay(
                            rayStart,
                            { x: rayDirX, y: 0, z: rayDirZ },
                            { maxDistance: 1, includePassableBlocks: false, includeLiquidBlocks: false }
                        );

                        if (rayResult?.block && !rayResult.block.isAir && !rayResult.block.isLiquid) {
                            const hitDistance = Math.sqrt(
                                Math.pow(rayResult.block.location.x + rayResult.faceLocation.x - location.x, 2) +
                                Math.pow(rayResult.block.location.z + rayResult.faceLocation.z - location.z, 2)
                            );

                            if (hitDistance <= 0.4) {
                                result.walls[side] = {
                                    block: rayResult.block,
                                    distance: hitDistance,
                                    angle: (angleOffset * 180 / Math.PI).toFixed(1) // Store angle for debugging
                                };
                                break;
                            }
                        }
                    }
                    if (result.walls[side]) break; // Stop checking angles once a wall is found
                }
            }
        }

        // Check for a block gap (for crawl/slide)
        if (opts.checkGap) {
            const xOffset = opts.usePlayerForward ? Math.cos(yawRad) * this.FORWARD_DISTANCE : 0;
            const zOffset = opts.usePlayerForward ? Math.sin(yawRad) * this.FORWARD_DISTANCE : 0;

            for (const dx of offsets) {
                for (const dz of offsets) {
                    const scanLoc = {
                        x: location.x + xOffset + dx,
                        y: location.y + 0.1,
                        z: location.z + zOffset + dz
                    };

                    const upRay = dimension.getBlockFromRay(
                        scanLoc, 
                        { x: 0, y: 1, z: 0 }, 
                        { maxDistance: 1, includePassableBlocks: false, includeLiquidBlocks: false }
                    );
                    const isHeadSolid = upRay && !upRay.block.isAir && !upRay.block.isLiquid;

                    const downRay = dimension.getBlockFromRay(
                        scanLoc, 
                        { x: 0, y: 0, z: 0 }, 
                        { maxDistance: 1, includePassableBlocks: false, includeLiquidBlocks: false }
                    );
                    const isFeetAir = !downRay || downRay.block.isAir || downRay.block.isLiquid;

                    if (isHeadSolid && isFeetAir) {
                        result.blockGap.exists = true;
                        result.blockGap.location = scanLoc;
                        break;
                    }
                }
                if (result.blockGap.exists) break;
            }
        }

        // Cache the result
        this.scanCache.set(cacheKey, { result, timestamp: currentTick });
        return result;
    }

    /**
     * Convenience method to check if a player can crawl through a gap
     * @param {Player} player - The player to check
     * @param {boolean} useForwardOffset - Whether to check ahead of the player
     * @returns {boolean} Whether a crawlable gap exists
     */
    checkBlockGap(player, useForwardOffset = false) {
        const result = this.scan(player, { 
            checkGround: false, 
            checkObstacles: false, 
            checkGap: true,
            usePlayerForward: useForwardOffset
        });
        return result.blockGap.exists;
    }

    /**
     * Specialized gap check for roll mechanics that matches the original implementation
     * @param {Player} player - The player to check
     * @param {boolean} useForwardOffset - Whether to check ahead of the player
     * @returns {boolean} Whether a rollable gap exists
     */
    checkRollGap(player, useForwardOffset) {
        const { location, dimension } = player;
        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;
        const offsets = [-0.29, 0, 0.29];
        const xOffset = useForwardOffset ? Math.cos(yawRad) * this.FORWARD_DISTANCE : 0;
        const zOffset = useForwardOffset ? Math.sin(yawRad) * this.FORWARD_DISTANCE : 0;

        for (const dx of offsets) {
            for (const dz of offsets) {
                const scanLoc = {
                    x: location.x + xOffset + dx,
                    y: location.y + 0.6,
                    z: location.z + zOffset + dz
                };

                const upRay = dimension.getBlockFromRay(scanLoc, { x: 0, y: 1, z: 0 }, { maxDistance: 1, includePassableBlocks: false, includeLiquidBlocks: false });
                const isHeadSolid = upRay && !upRay.block.isAir && !upRay.block.isLiquid;

                const downRay = dimension.getBlockFromRay(scanLoc, { x: 0, y: 0, z: 0 }, { maxDistance: 1, includePassableBlocks: false, includeLiquidBlocks: false });
                const isFeetAir = !downRay || downRay.block.isAir || downRay.block.isLiquid;

                if (isHeadSolid && isFeetAir) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Convenience method to detect a wall for wall running
     * @param {Player} player - The player to check
     * @returns {string|null} The side of the wall ("left", "right") or null if no wall
     */
    detectWall(player) {
        const result = this.scan(player, { 
            checkGround: false, 
            checkObstacles: false, 
            checkWalls: true 
        });
        
        if (result.walls.left) return "left";
        if (result.walls.right) return "right";
        return null;
    }

    /**
     * Convenience method to check vault conditions
     * @param {Player} player - The player to check
     * @returns {Object} Vault validation data
     */
    checkVaultBlock(player) {
        const { location, dimension } = player;
        const viewDir = player.getViewDirection();
        const { x, y, z } = location;
        const minVaultHeight = y + 0.9; // 0.9 block above feet level
        const maxVaultHeight = y + 1.5; // 1.5 blocks above feet level
        const scanLimit = y + 3; // Scan up to max vault height
        const heightAboveGround = this.getHeightAboveGround(player);
        
        let foundSolid = false;
        let tooHigh = false;
        let tooLow = true;
        let isSpecialBlock = false;
        
        const rayOptions = {
            maxDistance: 1,
            includePassableBlocks: false,
            includeLiquidBlocks: false
        };
        
        for (let scanY = y; scanY <= scanLimit; scanY += 0.1) {
            const rayStart = { x, y: scanY, z };
            const rayEnd = { 
                x: viewDir.x, 
                y: 0, // Horizontal ray for vault detection
                z: viewDir.z 
            };
            
            const rayResult = dimension.getBlockFromRay(rayStart, rayEnd, rayOptions);
            
            if (rayResult?.block && !rayResult.block.isAir && !rayResult.block.isLiquid) {
                foundSolid = true;
                const blockType = rayResult.block.typeId;
                // Check if block is a fence, wall, or gate
                isSpecialBlock = blockType.includes('fence') || 
                                blockType.includes('wall') || 
                                blockType.includes('gate');
                
                // Check height constraints
                if (scanY > maxVaultHeight) {
                    tooHigh = true;
                    break;
                } else if (isSpecialBlock || heightAboveGround >= 1.5) {
                    tooLow = false; // Skip min height for special blocks or high ground
                } else if (scanY < minVaultHeight) {
                    tooLow = true;
                } else {
                    tooLow = false;
                }
            }
        }

        return {
            valid: foundSolid && !tooLow && !tooHigh,
            tooHigh,
            tooLow,
            isSpecialBlock
        };
    }

    /**
     * Get a player's height above the ground
     * @param {Player} player - The player to check
     * @returns {number} Distance to the ground
     */
    getHeightAboveGround(player) {
        const result = this.scan(player, { checkGround: true });
        return result.ground.distance;
    }

    /**
     * Clears the scan cache
     */
    clearCache() {
        this.scanCache.clear();
        this.comprehensiveCache.clear();
        this.lastComprehensiveScan.clear();
        clearBlockCache();
    }

    /**
     * Clear scan cache for a specific player
     * @param {string} playerName - Name of the player to clear cache for
     */
    clearPlayerCache(playerName) {
        this.comprehensiveCache.delete(playerName);
        this.lastComprehensiveScan.delete(playerName);
    }

    /**
     * Get cache statistics for debugging
     * @returns {Object} Cache statistics
     */
    getCacheStats() {
        return {
            cachedPlayers: this.comprehensiveCache.size,
            totalScans: this.lastComprehensiveScan.size,
            regularCacheSize: this.scanCache.size
        };
    }
}

// Export a singleton instance for better performance and shared caching
const blockScanner = new BlockScanner();
export default blockScanner;